{*******************************************************}
{                                                       }
{       Advanced SQL statement parser                   }
{       Copyright (c) 2001 AS Gaiasoft                  }
{       Created by Gert Kello                           }
{                                                       }
{*******************************************************}

unit gaAdvancedSQLParser;

interface

uses
  Classes, gaBasicSQLParser;

type
  TSQLStatementType = (sstSelect, sstInsert, sstUpdate, sstDelete,
    sstCreate, sstAlter, sstDrop, sstUnknown, sstNoStatementFound);

  TSQLStatementTypes = set of TSQLStatementType;

  TgaCustomSQLStatement = class;

  TgaSQLTokenObj = class (TObject)
  private
    FIsOriginal: Boolean;
    FQuoteChar: Char;
    FTokenQuoted: Boolean;
    FTokenString: string;
    FTokenType: TSQLToken;
    function GetTokenAsString: string;
  public
    procedure AssignTokenInfo(ASQLParser: TgaBasicSQLParser);
    procedure SetTokenInfo(const AString: string; ATokenType: TSQLToken; 
            AQuoted: Boolean; AQuoteChar: char);
    function TokenSymbolIs(const S: string): Boolean;
    property IsOriginal: Boolean read FIsOriginal write FIsOriginal;
    property QuoteChar: Char read FQuoteChar;
    property TokenAsString: string read GetTokenAsString;
    property TokenQuoted: Boolean read FTokenQuoted;
    property TokenString: string read FTokenString;
    property TokenType: TSQLToken read FTokenType;
  end;
  
  TgaTokenEvent = procedure (Sender: TObject; AToken: TgaSQLTokenObj) of object;
  TgaSQLTokenList = class (TList)
  private
    FOwnerStatement: TgaCustomSQLStatement;
    function GetItems(Index: Integer): TgaSQLTokenObj;
    procedure SetItems(Index: Integer; const Value: TgaSQLTokenObj);
  protected
    procedure DoAddToken(Sender: TObject; AToken: TgaSQLTokenObj);
    procedure DoAfterTokenAdd(AToken: TgaSQLTokenObj); virtual;
    procedure GetAllTokens(ATokenList: TgaSQLTokenList);
    function GetAsString: string; virtual;
    function GetTokenObjAsString(ATokenObj: TgaSQLTokenObj): string;
    property OwnerStatement: TgaCustomSQLStatement read FOwnerStatement;
  public
    constructor Create(AOwnerStatement: TgaCustomSQLStatement); virtual;
    function AddToken(AToken: TgaSQLTokenObj): Integer;
    function NewToken: TgaSQLTokenObj; virtual;
    property AsString: string read GetAsString;
    property Items[Index: Integer]: TgaSQLTokenObj read GetItems write SetItems;
            default;
  end;
  
  TgaListOfSQLTokenLists = class (TList)
  private
    FOwnsLists: Boolean;
    function GetItems(Index: Integer): TgaSQLTokenList;
    function GetLastItem: TgaSQLTokenList;
    procedure SetItems(Index: Integer; const Value: TgaSQLTokenList);
  protected
    procedure GetAllTokens(ATokenList: TgaSQLTokenList);
    function GetAsString: string; virtual;
    procedure Notify(Ptr: Pointer; Action: TListNotification); override;
  public
    constructor Create;
    procedure AddToken(AToken: TgaSQLTokenObj); virtual;
    property AsString: string read GetAsString;
    property Items[Index: Integer]: TgaSQLTokenList read GetItems write 
            SetItems; default;
    property LastItem: TgaSQLTokenList read GetLastItem;
    property OwnsLists: Boolean read FOwnsLists write FOwnsLists;
  end;
  
  TgaAdvancedSQLParser = class;

  TgaCustomSQLStatement = class (TObject)
  private
    FAllFields: TgaListOfSQLTokenLists;
    FAllTables: TgaListOfSQLTokenLists;
    FCurrentSQL: TgaSQLTokenList;
    FCurrentToken: TgaSQLTokenObj;
    FInternalStatementState: Integer;
    FOnTokenAdded: TgaTokenEvent;
    FOriginalSQL: TgaSQLTokenList;
    FOwnerParser: TgaAdvancedSQLParser;
    FStatusCode: Integer;
  protected
    procedure DoAfterStatementStateChange; virtual;
    procedure DoBeforeStatementStateChange(const NewStateOrd: LongInt); virtual;
    procedure DoStatementComplete; virtual;
    procedure DoTokenAdded(Sender: TObject; AToken: TgaSQLTokenObj); virtual;
    function GetAsString: string; virtual;
    function GetStatementType: TSQLStatementType; virtual; abstract;
    procedure ModifyStatementInErrorState(Sender: TObject; AToken: 
            TgaSQLTokenObj); virtual;
    procedure ModifyStatementInNormalState(Sender: TObject; AToken: 
            TgaSQLTokenObj); virtual;
    procedure SetCurrentSQL(Value: TgaSQLTokenList);
    procedure SetOriginalSQL(Value: TgaSQLTokenList);
    property CurrentToken: TgaSQLTokenObj read FCurrentToken write 
            FCurrentToken;
    property InternalStatementState: Integer read FInternalStatementState write 
            FInternalStatementState;
    property OwnerParser: TgaAdvancedSQLParser read FOwnerParser;
  public
    constructor Create(AOwner: TgaAdvancedSQLParser); virtual;
    destructor Destroy; override;
    procedure Clear; virtual;
    procedure DoTokenParsed;
    function InsertToken(AToken, AInsToken: TgaSQLTokenObj): Integer;
    function InsertTokenAfter(AToken, AInsToken: TgaSQLTokenObj): Integer;
    procedure MoveDataTo(Destination: TgaCustomSQLStatement); virtual;
    procedure RemoveToken(AToken: TgaSQLTokenObj);
    procedure RemoveTokenList(AToken: TgaSQLTokenList);
    procedure ReplaceToken(AOldToken, ANewToken: TgaSQLTokenObj);
    procedure ReplaceTokenList(AInsertTokenList, ARemoveTokenList: 
            TgaSQLTokenList);
    property AllFields: TgaListOfSQLTokenLists read FAllFields;
    property AllTables: TgaListOfSQLTokenLists read FAllTables;
    property AsString: string read GetAsString;
    property CurrentSQL: TgaSQLTokenList read FCurrentSQL;
    property OnTokenAdded: TgaTokenEvent read FOnTokenAdded write FOnTokenAdded;
    property OriginalSQL: TgaSQLTokenList read FOriginalSQL;
    property StatementType: TSQLStatementType read GetStatementType;
    property StatusCode: Integer read FStatusCode write FStatusCode;
  end;
  
  TgaSQLSTatementClass = class of TgaCustomSQLStatement;

  TgaAdvancedSQLParser = class (TgaBasicSQLParser)
  private
    FCurrentStatement: TgaCustomSQLStatement;
    FOnStatementComplete: TNotifyEvent;
    function GetCurrentStatement: TgaCustomSQLStatement;
  protected
    procedure DoStatementComplete; virtual;
    procedure DoTokenParsed; override;
    function GetStatementClass: TgaSQLSTatementClass;
    procedure SetCurrentStatement(AStatement: TgaCustomSQLStatement);
  public
    destructor Destroy; override;
    class function AddStatementClass(const ATokenSymbol: string; 
            AStatementClass: TgaSQLSTatementClass): Integer;
    class function GetStatementClassForToken(const ATokenSymbol: string): 
            TgaSQLSTatementClass;
    class procedure RemoveStatementClass(const ATokenSymbol: string; 
            AStatementClass: TgaSQLSTatementClass);
    procedure Reset; override;
    property CurrentStatement: TgaCustomSQLStatement read GetCurrentStatement;
    property OnStatementComplete: TNotifyEvent read FOnStatementComplete write 
            FOnStatementComplete;
  end;
  
  TgaUnkownSQLStatement = class (TgaCustomSQLStatement)
  protected
    function GetStatementType: TSQLStatementType; override;
  end;
  
  TgaNoSQLStatement = class (TgaCustomSQLStatement)
  protected
    function GetStatementType: TSQLStatementType; override;
    procedure ModifyStatementInNormalState(Sender: TObject; AToken: 
            TgaSQLTokenObj); override;
  end;
  
const
  DMLStatementTypes = [sstSelect, sstInsert, sstUpdate, sstDelete];
  DDLStatementTypes = [sstCreate, sstAlter, sstDrop];

  errWrongKeywordSequence = $101;
  errUnexpectedTokenInStatement = $102;

procedure TrimTokenList(ATokenList: TgaSQLTokenList;
    const FreeRemovedTokens: boolean;
    TrimmedTokenTypes: TSQLTokenTypes = [stDelimitier, stEnd]);

implementation

uses
  SysUtils, gaSelectStm, gaUpdateStm, gaDeleteStm, gaInsertStm,
  gaSQLParserConsts;

// START resource string wizard section
resourcestring
  SInvalidTokenListToBeReplaced = 'Invalid token list to be replaced in %s.%s';

// END resource string wizard section


var
  StatementClassList: TStrings;

procedure TrimTokenList(ATokenList: TgaSQLTokenList;
    const FreeRemovedTokens: boolean;
    TrimmedTokenTypes: TSQLTokenTypes = [stDelimitier, stEnd]);
begin
  while (ATokenList.Count > 0) and (ATokenList.Items[ATokenList.Count - 1].TokenType in TrimmedTokenTypes) do
  begin
    if FreeRemovedTokens then
      ATokenList.Items[ATokenList.Count - 1].Free;
    ATokenList.Delete(ATokenList.Count - 1);
  end;
  while (ATokenList.Count > 0) and (ATokenList.Items[0].TokenType in TrimmedTokenTypes) do
  begin
    if FreeRemovedTokens then
      ATokenList.Items[ATokenList.Count - 1].Free;
    ATokenList.Delete(0);
  end;
end;

{
***************************** TgaAdvancedSQLParser *****************************
}
destructor TgaAdvancedSQLParser.Destroy;
begin
  FCurrentStatement.Free;
  inherited;
end;

class function TgaAdvancedSQLParser.AddStatementClass(const ATokenSymbol: 
        string; AStatementClass: TgaSQLSTatementClass): Integer;
begin
  Result := StatementClassList.AddObject(UpperCase(ATokenSymbol), TObject(AStatementClass));
end;

procedure TgaAdvancedSQLParser.DoStatementComplete;
begin
  if Assigned(FOnStatementComplete) then
    FOnStatementComplete(Self);
end;

procedure TgaAdvancedSQLParser.DoTokenParsed;
begin
  CurrentStatement.DoTokenParsed;
  inherited;
end;

function TgaAdvancedSQLParser.GetCurrentStatement: TgaCustomSQLStatement;
begin
  if not Assigned(FCurrentStatement) then
    FCurrentStatement := GetStatementClass.Create(Self);
  Result := FCurrentStatement;
end;

function TgaAdvancedSQLParser.GetStatementClass: TgaSQLSTatementClass;
begin
  Result := TgaNoSQLStatement;
  if TokenType = stSymbol then
    Result := GetStatementClassForToken(TokenString);
end;

class function TgaAdvancedSQLParser.GetStatementClassForToken(const 
        ATokenSymbol: string): TgaSQLSTatementClass;
var
  i: Integer;
  lTokenSymbol: string;
begin
  Result := TgaUnkownSQLSTatement;
  lTokenSymbol := UpperCase(ATokenSymbol);
  for i := StatementClassList.Count - 1 downto 0 do
    if StatementClassList[i] = lTokenSymbol then
    begin
      Result := TgaSQLSTatementClass(StatementClassList.Objects[i]);
      Exit;
    end;
end;

class procedure TgaAdvancedSQLParser.RemoveStatementClass(const ATokenSymbol: 
        string; AStatementClass: TgaSQLSTatementClass);
var
  i: Integer;
  lTokenSymbol: string;
begin
  lTokenSymbol := UpperCase(ATokenSymbol);
  for i := StatementClassList.Count - 1 downto 0 do
    if (StatementClassList[i] = lTokenSymbol) and
      (StatementClassList.Objects[i] = TObject(AStatementClass)) then
    begin
      StatementClassList.Delete(i);
      Exit;
    end;
end;

procedure TgaAdvancedSQLParser.Reset;
begin
  FreeAndNil(FCurrentStatement);
  inherited;
end;

procedure TgaAdvancedSQLParser.SetCurrentStatement(AStatement: 
        TgaCustomSQLStatement);
begin
  FCurrentStatement := AStatement;
end;

{
**************************** TgaUnkownSQLStatement *****************************
}
function TgaUnkownSQLStatement.GetStatementType: TSQLStatementType;
begin
  Result := sstUnknown;
end;

{
****************************** TgaNoSQLStatement *******************************
}
function TgaNoSQLStatement.GetStatementType: TSQLStatementType;
begin
  Result := sstNoStatementFound;
end;

procedure TgaNoSQLStatement.ModifyStatementInNormalState(Sender: TObject; 
        AToken: TgaSQLTokenObj);
var
  tmpSQLStatement: TgaCustomSQLStatement;
begin
  if OwnerParser.GetStatementClass <> TgaNoSQLStatement then
  begin
    tmpSQLStatement := OwnerParser.GetStatementClass.Create(OwnerParser);
    OwnerParser.SetCurrentStatement(tmpSQLStatement);
    MoveDataTo(tmpSQLStatement);
    tmpSQLStatement.ModifyStatementInNormalState(Sender, AToken);
    Self.Free;
  end else
    inherited;
end;

{
**************************** TgaCustomSQLStatement *****************************
}
constructor TgaCustomSQLStatement.Create(AOwner: TgaAdvancedSQLParser);
begin
  inherited Create;
  FOwnerParser := AOwner;
  FStatusCode := 0;
  FOriginalSQL := TgaSQLTokenList.Create(Self);
  FCurrentSQL := TgaSQLTokenList.Create(Self);
  FAllTables := TgaListOfSQLTokenLists.Create;
  FAllTables.OwnsLists := False;
  FAllFields := TgaListOfSQLTokenLists.Create;
  FAllFields.OwnsLists := False;
  FOnTokenAdded := ModifyStatementInNormalState;
end;

destructor TgaCustomSQLStatement.Destroy;
begin
  Clear;
  FOriginalSQL.Free;
  FCurrentSQL.Free;
  FAllTables.Free;
  FAllFields.Free;
  inherited;
end;

procedure TgaCustomSQLStatement.Clear;
var
  i: Integer;
begin
  if Assigned(CurrentSQL) then
    with CurrentSQL do
    begin
      for i := Count - 1 downto 0 do
        if not Items[i].IsOriginal then
          Items[i].Free;
      Clear;
    end;
  if Assigned(OriginalSQL) then
    with OriginalSQL do
    begin
      for i := Count - 1 downto 0 do
        Items[i].Free;
      Clear;
    end;
  AllTables.Clear;
  AllFields.Clear;
end;

procedure TgaCustomSQLStatement.DoAfterStatementStateChange;
begin
  // do nothing procedure
end;

procedure TgaCustomSQLStatement.DoBeforeStatementStateChange(const NewStateOrd: 
        LongInt);
begin
  // do nothing procedure
end;

procedure TgaCustomSQLStatement.DoStatementComplete;
begin
  if Assigned(FOwnerParser) then
    FOwnerParser.DoStatementComplete;
end;

procedure TgaCustomSQLStatement.DoTokenAdded(Sender: TObject; AToken: 
        TgaSQLTokenObj);
begin
  if Assigned(FOnTokenAdded) then FOnTokenAdded(Sender, AToken);
end;

procedure TgaCustomSQLStatement.DoTokenParsed;
begin
  CurrentToken := OriginalSQL.NewToken;
  CurrentToken.AssignTokenInfo(OwnerParser);
  CurrentToken.IsOriginal := True;
  CurrentSQL.AddToken(CurrentToken);
  DoTokenAdded(Self, CurrentToken);
  if StatusCode <> 0 then
    OnTokenAdded := ModifyStatementInErrorState;
end;

function TgaCustomSQLStatement.GetAsString: string;
begin
  Result := CurrentSQL.AsString;
end;

function TgaCustomSQLStatement.InsertToken(AToken, AInsToken: TgaSQLTokenObj): 
        Integer;
var
  lTokenIndex: Integer;
begin
  lTokenIndex := CurrentSQL.IndexOf(AInsToken);
  if lTokenIndex < 0 then
    raise Exception.Create(STokenInsertBeforeNotInSQL);
  CurrentSQL.Insert(lTokenIndex, AToken);
  Result := lTokenIndex;
end;

function TgaCustomSQLStatement.InsertTokenAfter(AToken, AInsToken: 
        TgaSQLTokenObj): Integer;
var
  lTokenIndex: Integer;
begin
  lTokenIndex := CurrentSQL.IndexOf(AInsToken);
  if lTokenIndex < 0 then
    raise Exception.Create(STokenInsertBeforeNotInSQL);
  CurrentSQL.Insert(lTokenIndex + 1, AToken);
  Result := lTokenIndex;
end;

procedure TgaCustomSQLStatement.ModifyStatementInErrorState(Sender: TObject; 
        AToken: TgaSQLTokenObj);
begin
  // Do nothing here
end;

procedure TgaCustomSQLStatement.ModifyStatementInNormalState(Sender: TObject; 
        AToken: TgaSQLTokenObj);
begin
  // Do nothing here
end;

procedure TgaCustomSQLStatement.MoveDataTo(Destination: TgaCustomSQLStatement);
begin
  Destination.SetOriginalSQL(OriginalSQL);
  FOriginalSQL := nil;
  Destination.SetCurrentSQL(CurrentSQL);
  FCurrentSQL := nil;
  Destination.CurrentToken := CurrentToken;
  FCurrentToken := nil;
end;

procedure TgaCustomSQLStatement.RemoveToken(AToken: TgaSQLTokenObj);
var
  lTokenIndex: Integer;
begin
  lTokenIndex := CurrentSQL.IndexOf(AToken);
  if lTokenIndex < 0 then
    raise Exception.Create(STokenToBeRemovedNotInSQL);
  CurrentSQL.Delete(lTokenIndex);
  if not AToken.IsOriginal then
    AToken.Free;
end;

procedure TgaCustomSQLStatement.RemoveTokenList(AToken: TgaSQLTokenList);
begin
end;

procedure TgaCustomSQLStatement.ReplaceToken(AOldToken, ANewToken: 
        TgaSQLTokenObj);
var
  lTokenIndex: Integer;
begin
  lTokenIndex := CurrentSQL.IndexOf(AOldToken);
  if lTokenIndex < 0 then
    raise Exception.Create(STokenToBeReplacedNotInSQL);
  CurrentSQL.Items[lTokenIndex] := ANewToken;
  if not AOldToken.IsOriginal then
    AOldToken.Free;
end;

procedure TgaCustomSQLStatement.ReplaceTokenList(AInsertTokenList, 
        ARemoveTokenList: TgaSQLTokenList);
var
  lInsertPoint: Integer;
  lTokenIndex: Integer;
  i: Integer;
begin
  lInsertPoint := CurrentSQL.IndexOf(ARemoveTokenList[0]);
  if lInsertPoint < 0 then
    raise Exception.CreateFmt(SInvalidTokenListToBeReplaced, [ClassName, 'ReplaceTokenList']);
  for i := ARemoveTokenList.Count - 1 downto 0 do
  begin
      lTokenIndex := CurrentSQL.IndexOf(ARemoveTokenList[i]);
      if lTokenIndex <> lInsertPoint + i then
        raise Exception.CreateFmt(SInvalidTokenListToBeReplaced, [ClassName, 'ReplaceTokenList']);
      CurrentSQL.Delete(lTokenIndex);
      if not ARemoveTokenList[i].IsOriginal then
      begin
        ARemoveTokenList[i].Free;
        ARemoveTokenList.Delete(i);
      end;
  end;
  for i := AInsertTokenList.Count - 1 downto 0 do
    CurrentSQL.Insert(lInsertPoint, AInsertTokenList[i]);
end;

procedure TgaCustomSQLStatement.SetCurrentSQL(Value: TgaSQLTokenList);
begin
  FCurrentSQL.Free;
  FCurrentSQL := Value;
end;

procedure TgaCustomSQLStatement.SetOriginalSQL(Value: TgaSQLTokenList);
begin
  FOriginalSQL.Free;
  FOriginalSQL := Value;
end;

{
******************************* TgaSQLTokenList ********************************
}
constructor TgaSQLTokenList.Create(AOwnerStatement: TgaCustomSQLStatement);
begin
  inherited Create;
  FOwnerStatement := AOwnerStatement;
end;

function TgaSQLTokenList.AddToken(AToken: TgaSQLTokenObj): Integer;
begin
  Result := inherited Add(AToken);
  DoAfterTokenAdd(AToken);
end;

procedure TgaSQLTokenList.DoAddToken(Sender: TObject; AToken: TgaSQLTokenObj);
begin
  AddToken(AToken);
end;

procedure TgaSQLTokenList.DoAfterTokenAdd(AToken: TgaSQLTokenObj);
begin
  // Do nothing here
end;

procedure TgaSQLTokenList.GetAllTokens(ATokenList: TgaSQLTokenList);
var
  i: Integer;
begin
  ATokenList.Clear;
  for i := 0 to Count - 1 do
    ATokenList.Add(Items[i]);
end;

function TgaSQLTokenList.GetAsString: string;
var
  i: Integer;
begin
  Result := EmptyStr;
  for i := 0 to Count - 1 do
    Result := Result + Items[i].TokenAsString;
end;

function TgaSQLTokenList.GetItems(Index: Integer): TgaSQLTokenObj;
begin
  Result := TgaSQLTokenObj(inherited Items[Index]);
end;

function TgaSQLTokenList.GetTokenObjAsString(ATokenObj: TgaSQLTokenObj): string;
begin
  if Assigned(ATokenObj) then
    Result := ATokenObj.TokenAsString
  else
    Result := EmptyStr;
end;

function TgaSQLTokenList.NewToken: TgaSQLTokenObj;
begin
  Result := TgaSQLTokenObj.Create;
  AddToken(Result);
end;

procedure TgaSQLTokenList.SetItems(Index: Integer; const Value: TgaSQLTokenObj);
begin
  inherited Items[Index] := Value;
end;

{
******************************** TgaSQLTokenObj ********************************
}
procedure TgaSQLTokenObj.AssignTokenInfo(ASQLParser: TgaBasicSQLParser);
begin
  FTokenType := ASQLParser.TokenType;
  FTokenString := ASQLParser.TokenString;
  FQuoteChar := ASQLParser.QuoteChar;
  FTokenQuoted := ASQLParser.TokenQuoted;
end;

function TgaSQLTokenObj.GetTokenAsString: string;
begin
  Result := TokenString;
  if TokenQuoted then
    Result := QuoteChar + Result + QuoteChar;
  if TokenType = stParameter then
    Result := ':' + Result;
end;

procedure TgaSQLTokenObj.SetTokenInfo(const AString: string; ATokenType: 
        TSQLToken; AQuoted: Boolean; AQuoteChar: char);
begin
  FTokenType := ATokenType;
  FTokenString := AString;
  FQuoteChar := AQuoteChar;
  FTokenQuoted := AQuoted;
end;

function TgaSQLTokenObj.TokenSymbolIs(const S: string): Boolean;
begin
  Result := (TokenType = stSymbol) and (CompareText(FTokenString, S) = 0);
end;

{
**************************** TgaListOfSQLTokenLists ****************************
}
constructor TgaListOfSQLTokenLists.Create;
begin
  inherited;
  FOwnsLists := True;
end;

procedure TgaListOfSQLTokenLists.AddToken(AToken: TgaSQLTokenObj);
begin
  raise Exception.Create('Can''t add token to the list of token lists. Add it to the token list instead');
end;

procedure TgaListOfSQLTokenLists.GetAllTokens(ATokenList: TgaSQLTokenList);
var
  i: Integer;
  ii: Integer;
  AItem: TgaSQLTokenList;
begin
  ATokenList.Clear;
  for i := 0 to Count - 1 do
  begin
    AItem := Items[i];
    for ii := 0 to AItem.Count - 1 do
      ATokenList.Add(AItem[ii]);
  end;
end;

function TgaListOfSQLTokenLists.GetAsString: string;
var
  i: Integer;
begin
  Result := EmptyStr;
  for i := 0 to Count - 1 do
    Result := Result + Items[i].AsString;
end;

function TgaListOfSQLTokenLists.GetItems(Index: Integer): TgaSQLTokenList;
begin
  Result := TgaSQLTokenList(inherited Items[Index]);
end;

function TgaListOfSQLTokenLists.GetLastItem: TgaSQLTokenList;
begin
  Result := Items[Count - 1];
end;

procedure TgaListOfSQLTokenLists.Notify(Ptr: Pointer; Action: 
        TListNotification);
begin
  if OwnsLists then
    if Action = lnDeleted then
      TObject(Ptr).Free;
  inherited;
end;

procedure TgaListOfSQLTokenLists.SetItems(Index: Integer; const Value: 
        TgaSQLTokenList);
begin
  inherited Items[Index] := Value;
end;

initialization
  StatementClassList := TStringList.Create;
  TgaAdvancedSQLParser.AddStatementClass('SELECT', TgaSelectSQLStatement);
  TgaAdvancedSQLParser.AddStatementClass('UPDATE', TgaUpdateSQLStatement);
  TgaAdvancedSQLParser.AddStatementClass('DELETE', TgaDeleteSQLStatement);
  TgaAdvancedSQLParser.AddStatementClass('INSERT', TgaInsertSQLStatement);

finalization
  TgaAdvancedSQLParser.RemoveStatementClass('SELECT', TgaSelectSQLStatement);
  TgaAdvancedSQLParser.RemoveStatementClass('UPDATE', TgaUpdateSQLStatement);
  TgaAdvancedSQLParser.RemoveStatementClass('DELETE', TgaDeleteSQLStatement);
  TgaAdvancedSQLParser.RemoveStatementClass('INSERT', TgaInsertSQLStatement);
  StatementClassList.Free;

end.
